"""
    flexparser.flexparser
    ~~~~~~~~~~~~~~~~~~~~~

    Classes and functions to create parsers.

    The idea is quite simple. You write a class for every type of content
    (called here ``ParsedStatement``) you need to parse. Each class should
    have a ``from_string`` constructor. We used extensively the ``typing``
    module to make the output structure easy to use and less error prone.

    For more information, take a look at https://github.com/hgrecco/flexparser

    :copyright: 2022 by flexparser Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import sys
import collections
import dataclasses
import enum
import functools
import hashlib
import hmac
import inspect
import logging
import pathlib
import re
import typing as ty
from dataclasses import dataclass
from functools import cached_property
from importlib import resources
from typing import Any, Union, Optional, no_type_check

if sys.version_info >= (3, 10):
    from typing import TypeAlias  # noqa
else:
    from typing_extensions import TypeAlias  # noqa


if sys.version_info >= (3, 11):
    from typing import Self  # noqa
else:
    from typing_extensions import Self  # noqa


_LOGGER = logging.getLogger("flexparser")

_SENTINEL = object()


class HasherProtocol(ty.Protocol):
    @property
    def name(self) -> str:
        ...

    def hexdigest(self) -> str:
        ...


class GenericInfo:
    _specialized: Optional[
        dict[type, Optional[list[tuple[type, dict[ty.TypeVar, type]]]]]
    ] = None

    @staticmethod
    def _summarize(d: dict[ty.TypeVar, type]) -> dict[ty.TypeVar, type]:
        d = d.copy()
        while True:
            for k, v in d.items():
                if isinstance(v, ty.TypeVar):
                    d[k] = d[v]
                    break
            else:
                return d

            del d[v]

    @classmethod
    def _specialization(cls) -> dict[ty.TypeVar, type]:
        if cls._specialized is None:
            return dict()

        out: dict[ty.TypeVar, type] = {}
        specialized = cls._specialized[cls]

        if specialized is None:
            return {}

        for parent, content in specialized:
            for tvar, typ in content.items():
                out[tvar] = typ
                origin = getattr(parent, "__origin__", None)
                if origin is not None and origin in cls._specialized:
                    out = {**origin._specialization(), **out}

        return out

    @classmethod
    def specialization(cls) -> dict[ty.TypeVar, type]:
        return GenericInfo._summarize(cls._specialization())

    def __init_subclass__(cls) -> None:
        if cls._specialized is None:
            cls._specialized = {GenericInfo: None}

        tv: list[ty.TypeVar] = []
        entries: list[tuple[type, dict[ty.TypeVar, type]]] = []

        for par in getattr(cls, "__parameters__", ()):
            if isinstance(par, ty.TypeVar):
                tv.append(par)

        for b in getattr(cls, "__orig_bases__", ()):
            for k in cls._specialized.keys():
                if getattr(b, "__origin__", None) is k:
                    entries.append((b, {k: v for k, v in zip(tv, b.__args__)}))
                    break

        cls._specialized[cls] = entries

        return super().__init_subclass__()


################
# Exceptions
################


@dataclass(frozen=True)
class Statement:
    """Base class for parsed elements within a source file."""

    is_position_set: bool = dataclasses.field(init=False, default=False, repr=False)

    start_line: int = dataclasses.field(init=False, default=0)
    start_col: int = dataclasses.field(init=False, default=0)

    end_line: int = dataclasses.field(init=False, default=0)
    end_col: int = dataclasses.field(init=False, default=0)

    raw: Optional[str] = dataclasses.field(init=False, default=None)

    @classmethod
    def from_statement(cls, statement: Statement) -> Self:
        out = cls()
        if statement.is_position_set:
            out.set_position(*statement.get_position())
        if statement.raw is not None:
            out.set_raw(statement.raw)
        return out

    @classmethod
    def from_statement_iterator_element(
        cls, values: tuple[int, int, int, int, str]
    ) -> Self:
        out = cls()
        out.set_position(*values[:-1])
        out.set_raw(values[-1])
        return out

    @property
    def format_position(self) -> str:
        if not self.is_position_set:
            return "N/A"
        return "%d,%d-%d,%d" % self.get_position()

    @property
    def raw_strip(self) -> Optional[str]:
        if self.raw is None:
            return None
        return self.raw.strip()

    def get_position(self) -> tuple[int, int, int, int]:
        if self.is_position_set:
            return self.start_line, self.start_col, self.end_line, self.end_col
        return 0, 0, 0, 0

    def set_position(
        self: Self, start_line: int, start_col: int, end_line: int, end_col: int
    ) -> Self:
        object.__setattr__(self, "is_position_set", True)
        object.__setattr__(self, "start_line", start_line)
        object.__setattr__(self, "start_col", start_col)
        object.__setattr__(self, "end_line", end_line)
        object.__setattr__(self, "end_col", end_col)
        return self

    def set_raw(self: Self, raw: str) -> Self:
        object.__setattr__(self, "raw", raw)
        return self

    def set_simple_position(self: Self, line: int, col: int, width: int) -> Self:
        return self.set_position(line, col, line, col + width)


@dataclass(frozen=True)
class ParsingError(Statement, Exception):
    """Base class for all parsing exceptions in this package."""

    def __str__(self) -> str:
        return Statement.__str__(self)


@dataclass(frozen=True)
class UnknownStatement(ParsingError):
    """A string statement could not bee parsed."""

    def __str__(self) -> str:
        return f"Could not parse '{self.raw}' ({self.format_position})"


@dataclass(frozen=True)
class UnhandledParsingError(ParsingError):
    """Base class for all parsing exceptions in this package."""

    ex: Exception

    def __str__(self) -> str:
        return f"Unhandled exception while parsing '{self.raw}' ({self.format_position}): {self.ex}"


@dataclass(frozen=True)
class UnexpectedEOS(ParsingError):
    """End of file was found within an open block."""


#############################
# Useful methods and classes
#############################


@dataclass(frozen=True)
class Hash:
    algorithm_name: str
    hexdigest: str

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Hash)
            and self.algorithm_name != ""
            and self.algorithm_name == other.algorithm_name
            and hmac.compare_digest(self.hexdigest, other.hexdigest)
        )

    @classmethod
    def from_bytes(
        cls,
        algorithm: ty.Callable[
            [
                bytes,
            ],
            HasherProtocol,
        ],
        b: bytes,
    ) -> Self:
        hasher = algorithm(b)
        return cls(hasher.name, hasher.hexdigest())

    @classmethod
    def from_file_pointer(
        cls,
        algorithm: ty.Callable[
            [
                bytes,
            ],
            HasherProtocol,
        ],
        fp: ty.BinaryIO,
    ) -> Self:
        return cls.from_bytes(algorithm, fp.read())

    @classmethod
    def nullhash(cls) -> Self:
        return cls("", "")


def _yield_types(
    obj: type,
    valid_subclasses: tuple[type, ...] = (object,),
    recurse_origin: tuple[Any, ...] = (tuple, list, Union),
) -> ty.Generator[type, None, None]:
    """Recursively transverse type annotation if the
    origin is any of the types in `recurse_origin`
    and yield those type which are subclasses of `valid_subclasses`.

    """
    if ty.get_origin(obj) in recurse_origin:
        for el in ty.get_args(obj):
            yield from _yield_types(el, valid_subclasses, recurse_origin)
    else:
        if inspect.isclass(obj) and issubclass(obj, valid_subclasses):
            yield obj


class classproperty:  # noqa N801
    """Decorator for a class property

    In Python 3.9+ can be replaced by

        @classmethod
        @property
        def myprop(self):
            return 42

    """

    def __init__(self, fget):  # type: ignore
        self.fget = fget

    def __get__(self, owner_self, owner_cls):  # type: ignore
        return self.fget(owner_cls)  # type: ignore


class DelimiterInclude(enum.IntEnum):
    """Specifies how to deal with delimiters while parsing."""

    #: Split at delimiter, not including in any string
    SPLIT = enum.auto()

    #: Split after, keeping the delimiter with previous string.
    SPLIT_AFTER = enum.auto()

    #: Split before, keeping the delimiter with next string.
    SPLIT_BEFORE = enum.auto()

    #: Do not split at delimiter.
    DO_NOT_SPLIT = enum.auto()


class DelimiterAction(enum.IntEnum):
    """Specifies how to deal with delimiters while parsing."""

    #: Continue parsing normally.
    CONTINUE = enum.auto()

    #: Capture everything til end of line as a whole.
    CAPTURE_NEXT_TIL_EOL = enum.auto()

    #: Stop parsing line and move to next.
    STOP_PARSING_LINE = enum.auto()

    #: Stop parsing content.
    STOP_PARSING = enum.auto()


DO_NOT_SPLIT_EOL = {
    "\r\n": (DelimiterInclude.DO_NOT_SPLIT, DelimiterAction.CONTINUE),
    "\n": (DelimiterInclude.DO_NOT_SPLIT, DelimiterAction.CONTINUE),
    "\r": (DelimiterInclude.DO_NOT_SPLIT, DelimiterAction.CONTINUE),
}

SPLIT_EOL = {
    "\r\n": (DelimiterInclude.SPLIT, DelimiterAction.CONTINUE),
    "\n": (DelimiterInclude.SPLIT, DelimiterAction.CONTINUE),
    "\r": (DelimiterInclude.SPLIT, DelimiterAction.CONTINUE),
}

_EOLs_set = set(DO_NOT_SPLIT_EOL.keys())


@functools.lru_cache
def _build_delimiter_pattern(delimiters: tuple[str, ...]) -> re.Pattern[str]:
    """Compile a tuple of delimiters into a regex expression with a capture group
    around the delimiter.
    """
    return re.compile("|".join(f"({re.escape(el)})" for el in delimiters))


############
# Iterators
############

DelimiterDictT = dict[str, tuple[DelimiterInclude, DelimiterAction]]


class Spliter:
    """Content iterator splitting according to given delimiters.

    The pattern can be changed dynamically sending a new pattern to the ty.Generator,
    see DelimiterInclude and DelimiterAction for more information.

    The current scanning position can be changed at any time.

    Parameters
    ----------
    content : str
    delimiters : dict[str, tuple[DelimiterInclude, DelimiterAction]]

    Yields
    ------
    start_line : int
        line number of the start of the content (zero-based numbering).
    start_col : int
        column number of the start of the content (zero-based numbering).
    end_line : int
        line number of the end of the content (zero-based numbering).
    end_col : int
        column number of the end of the content (zero-based numbering).
    part : str
        part of the text between delimiters.
    """

    _pattern: Optional[re.Pattern[str]]
    _delimiters: DelimiterDictT

    __stop_searching_in_line: bool = False

    __pending: str = ""
    __first_line_col: Optional[tuple[int, int]] = None

    __lines: list[str]
    __lineno: int = 0
    __colno: int = 0

    def __init__(self, content: str, delimiters: DelimiterDictT):
        self.set_delimiters(delimiters)
        self.__lines = content.splitlines(keepends=True)

    def set_position(self, lineno: int, colno: int) -> None:
        self.__lineno, self.__colno = lineno, colno

    def set_delimiters(self, delimiters: DelimiterDictT) -> None:
        for k, v in delimiters.items():
            if v == (DelimiterInclude.DO_NOT_SPLIT, DelimiterAction.STOP_PARSING):
                raise ValueError(
                    f"The delimiter action for {k} is not a valid combination ({v})"
                )
        # Build a pattern but removing eols
        _pat_dlm = tuple(set(delimiters.keys()) - _EOLs_set)
        if _pat_dlm:
            self._pattern = _build_delimiter_pattern(_pat_dlm)
        else:
            self._pattern = None
        # We add the end of line as delimiters if not present.
        self._delimiters = {**DO_NOT_SPLIT_EOL, **delimiters}

    def __iter__(self) -> Spliter:
        return self

    def __next__(self) -> tuple[int, int, int, int, str]:
        if self.__lineno >= len(self.__lines):
            raise StopIteration

        while True:
            if self.__stop_searching_in_line:
                # There must be part of a line pending to parse
                # due to stop
                line = self.__lines[self.__lineno]
                mo = None
                self.__stop_searching_in_line = False
            else:
                # We get the current line and the find the first delimiter.
                line = self.__lines[self.__lineno]
                if self._pattern is None:
                    mo = None
                else:
                    mo = self._pattern.search(line, self.__colno)

            if mo is None:
                # No delimiter was found,
                # which should happen at end of the content or end of line
                for k in DO_NOT_SPLIT_EOL.keys():
                    if line.endswith(k):
                        dlm = line[-len(k) :]
                        end_col, next_col = len(line) - len(k), 0
                        break
                else:
                    # No EOL found, this is end of content
                    dlm = None
                    end_col, next_col = len(line), 0

                next_line = self.__lineno + 1

            else:
                next_line = self.__lineno
                end_col, next_col = mo.span()
                dlm = mo.group()

            part = line[self.__colno : end_col]

            if dlm is None:
                include, action = DelimiterInclude.SPLIT, DelimiterAction.STOP_PARSING
            else:
                include, action = self._delimiters[dlm]

            if include == DelimiterInclude.SPLIT:
                next_pending = ""
            else:
                # When dlm is None, DelimiterInclude.SPLIT
                assert isinstance(dlm, str)
                if include == DelimiterInclude.SPLIT_AFTER:
                    end_col += len(dlm)
                    part = part + dlm
                    next_pending = ""
                elif include == DelimiterInclude.SPLIT_BEFORE:
                    next_pending = dlm
                elif include == DelimiterInclude.DO_NOT_SPLIT:
                    self.__pending += line[self.__colno : end_col] + dlm
                    next_pending = ""
                else:
                    raise ValueError(f"Unknown action {include}.")

            if action == DelimiterAction.STOP_PARSING:
                # this will raise a StopIteration in the next call.
                next_line = len(self.__lines)
            elif action == DelimiterAction.STOP_PARSING_LINE:
                next_line = self.__lineno + 1
                next_col = 0

            start_line = self.__lineno
            start_col = self.__colno
            end_line = self.__lineno

            self.__lineno = next_line
            self.__colno = next_col

            if action == DelimiterAction.CAPTURE_NEXT_TIL_EOL:
                self.__stop_searching_in_line = True

            if include == DelimiterInclude.DO_NOT_SPLIT:
                self.__first_line_col = start_line, start_col
            else:
                if self.__first_line_col is None:
                    out = (
                        start_line,
                        start_col - len(self.__pending),
                        end_line,
                        end_col,
                        self.__pending + part,
                    )
                else:
                    out = (
                        *self.__first_line_col,
                        end_line,
                        end_col,
                        self.__pending + part,
                    )
                    self.__first_line_col = None
                self.__pending = next_pending
                return out


class StatementIterator:
    """Content peekable iterator splitting according to given delimiters.

    The pattern can be changed dynamically sending a new pattern to the ty.Generator,
    see DelimiterInclude and DelimiterAction for more information.

    Parameters
    ----------
    content : str
    delimiters : dict[str, tuple[DelimiterInclude, DelimiterAction]]

    Yields
    ------
    Statement
    """

    _cache: ty.Deque[Statement]

    def __init__(
        self, content: str, delimiters: DelimiterDictT, strip_spaces: bool = True
    ):
        self._cache = collections.deque()
        self._spliter = Spliter(content, delimiters)
        self._strip_spaces = strip_spaces

    def __iter__(self):
        return self

    def set_delimiters(self, delimiters: DelimiterDictT) -> None:
        self._spliter.set_delimiters(delimiters)
        if self._cache:
            value = self.peek()
            # Elements are 1 based indexing, while splitter is 0 based.
            self._spliter.set_position(value.start_line - 1, value.start_col)
            self._cache.clear()

    def _get_next_strip(self) -> Statement:
        part = ""
        while not part:
            start_line, start_col, end_line, end_col, part = next(self._spliter)
            lo = len(part)
            part = part.lstrip()
            start_col += lo - len(part)

            lo = len(part)
            part = part.rstrip()
            end_col -= lo - len(part)

        return Statement.from_statement_iterator_element(
            (start_line + 1, start_col, end_line + 1, end_col, part)  # type: ignore
        )

    def _get_next(self) -> Statement:
        if self._strip_spaces:
            return self._get_next_strip()

        part = ""
        while not part:
            start_line, start_col, end_line, end_col, part = next(self._spliter)

        return Statement.from_statement_iterator_element(
            (start_line + 1, start_col, end_line + 1, end_col, part)  # type: ignore
        )

    def peek(self, default: Any = _SENTINEL) -> Statement:
        """Return the item that will be next returned from ``next()``.

        Return ``default`` if there are no items left. If ``default`` is not
        provided, raise ``StopIteration``.

        """
        if not self._cache:
            try:
                self._cache.append(self._get_next())
            except StopIteration:
                if default is _SENTINEL:
                    raise
                return default
        return self._cache[0]

    def __next__(self) -> Statement:
        if self._cache:
            return self._cache.popleft()
        return self._get_next()


###########
# Parsing
###########

# Configuration type
T = ty.TypeVar("T")
CT = ty.TypeVar("CT")
PST = ty.TypeVar("PST", bound="ParsedStatement[Any]")
LineColStr: TypeAlias = tuple[int, int, str]

ParsedResult: TypeAlias = Union[T, ParsingError]
NullableParsedResult: TypeAlias = Union[T, ParsingError, None]


class ConsumeProtocol(ty.Protocol):
    @property
    def is_position_set(self) -> bool:
        ...

    @property
    def start_line(self) -> int:
        ...

    @property
    def start_col(self) -> int:
        ...

    @property
    def end_line(self) -> int:
        ...

    @property
    def end_col(self) -> int:
        ...

    @classmethod
    def consume(
        cls, statement_iterator: StatementIterator, config: Any
    ) -> NullableParsedResult[Self]:
        ...


@dataclass(frozen=True)
class ParsedStatement(ty.Generic[CT], Statement):
    """A single parsed statement.

    In order to write your own, you need to subclass it as a
    frozen dataclass and implement the parsing logic by overriding
    `from_string` classmethod.

    Takes two arguments: the string to parse and an object given
    by the parser which can be used to store configuration information.

    It should return an instance of this class if parsing
    was successful or None otherwise
    """

    @classmethod
    def from_string(cls, s: str) -> NullableParsedResult[Self]:
        """Parse a string into a ParsedStatement.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class.
        2. A subclass of ParsedStatement: the string was parsed successfully
        3. A subclass of ParsingError the string could be parsed with this class but there is
           an error.
        """
        raise NotImplementedError(
            "ParsedStatement subclasses must implement "
            "'from_string' or 'from_string_and_config'"
        )

    @classmethod
    def from_string_and_config(cls, s: str, config: CT) -> NullableParsedResult[Self]:
        """Parse a string into a ParsedStatement.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class.
        2. A subclass of ParsedStatement: the string was parsed successfully
        3. A subclass of ParsingError the string could be parsed with this class but there is
           an error.
        """
        return cls.from_string(s)

    @classmethod
    def from_statement_and_config(
        cls, statement: Statement, config: CT
    ) -> NullableParsedResult[Self]:
        raw = statement.raw
        if raw is None:
            return None

        try:
            out = cls.from_string_and_config(raw, config)
        except Exception as ex:
            out = UnhandledParsingError(ex)

        if out is None:
            return None

        out.set_position(*statement.get_position())
        out.set_raw(raw)
        return out

    @classmethod
    def consume(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> NullableParsedResult[Self]:
        """Peek into the iterator and try to parse.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class, the iterator is kept an the current place.
        2. a subclass of ParsedStatement: the string was parsed successfully, advance the iterator.
        3. a subclass of ParsingError: the string could be parsed with this class but there is
           an error, advance the iterator.
        """
        statement = statement_iterator.peek()
        parsed_statement = cls.from_statement_and_config(statement, config)
        if parsed_statement is None:
            return None
        next(statement_iterator)
        return parsed_statement


OPST = ty.TypeVar("OPST", bound="ParsedStatement[Any]")
BPST = ty.TypeVar(
    "BPST", bound="Union[ParsedStatement[Any], Block[Any, Any, Any, Any]]"
)
CPST = ty.TypeVar("CPST", bound="ParsedStatement[Any]")
RBT = ty.TypeVar("RBT", bound="RootBlock[Any, Any]")


@dataclass(frozen=True)
class Block(ty.Generic[OPST, BPST, CPST, CT], GenericInfo):
    """A sequence of statements with an opening, body and closing."""

    opening: ParsedResult[OPST]
    body: tuple[ParsedResult[BPST], ...]
    closing: Union[ParsedResult[CPST], EOS[CT]]

    delimiters: DelimiterDictT = dataclasses.field(default_factory=dict, init=False)

    def is_closed(self) -> bool:
        return not isinstance(self.closing, EOS)

    @property
    def is_position_set(self) -> bool:
        return self.opening.is_position_set

    @property
    def start_line(self) -> int:
        return self.opening.start_line

    @property
    def start_col(self) -> int:
        return self.opening.start_col

    @property
    def end_line(self) -> int:
        return self.closing.end_line

    @property
    def end_col(self) -> int:
        return self.closing.end_col

    def get_position(self) -> tuple[int, int, int, int]:
        return self.start_line, self.start_col, self.end_line, self.end_col

    @property
    def format_position(self) -> str:
        if not self.is_position_set:
            return "N/A"
        return "%d,%d-%d,%d" % self.get_position()

    def __iter__(
        self,
    ) -> ty.Generator[
        ParsedResult[Union[OPST, BPST, Union[CPST, EOS[CT]]]], None, None
    ]:
        yield self.opening
        for el in self.body:
            if isinstance(el, Block):
                yield from el
            else:
                yield el
        yield self.closing

    def iter_blocks(
        self,
    ) -> ty.Generator[ParsedResult[Union[OPST, BPST, CPST]], None, None]:
        # raise RuntimeError("Is this used?")
        yield self.opening
        yield from self.body
        yield self.closing

    ###################################################
    # Convenience methods to iterate parsed statements
    ###################################################

    _ElementT = ty.TypeVar("_ElementT", bound=Statement)

    def filter_by(
        self, klass1: type[_ElementT], *klass: type[_ElementT]
    ) -> ty.Generator[_ElementT, None, None]:
        """Yield elements of a given class or classes."""
        yield from (el for el in self if isinstance(el, (klass1,) + klass))  # type: ignore[misc]

    @cached_property
    def errors(self) -> tuple[ParsingError, ...]:
        """Tuple of errors found."""
        return tuple(self.filter_by(ParsingError))

    @property
    def has_errors(self) -> bool:
        """True if errors were found during parsing."""
        return bool(self.errors)

    ####################
    # Statement classes
    ####################

    @classmethod
    def opening_classes(cls) -> ty.Generator[type[OPST], None, None]:
        """Classes representing any of the parsed statement that can open this block."""
        try:
            opening = cls.specialization()[OPST]  # type: ignore[misc]
        except KeyError:
            opening: type = ty.get_type_hints(cls)["opening"]  # type: ignore[no-redef]
        yield from _yield_types(opening, ParsedStatement)  # type: ignore

    @classmethod
    def body_classes(cls) -> ty.Generator[type[BPST], None, None]:
        """Classes representing any of the parsed statement that can be in the body."""
        try:
            body = cls.specialization()[BPST]  # type: ignore[misc]
        except KeyError:
            body: type = ty.get_type_hints(cls)["body"]  # type: ignore[no-redef]
        yield from _yield_types(body, (ParsedStatement, Block))  # type: ignore

    @classmethod
    def closing_classes(cls) -> ty.Generator[type[CPST], None, None]:
        """Classes representing any of the parsed statement that can close this block."""
        try:
            closing = cls.specialization()[CPST]  # type: ignore[misc]
        except KeyError:
            closing: type = ty.get_type_hints(cls)["closing"]  # type: ignore[no-redef]
        yield from _yield_types(closing, ParsedStatement)  # type: ignore

    ##########
    # ParsedResult
    ##########

    @classmethod
    def consume_opening(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> NullableParsedResult[OPST]:
        """Peek into the iterator and try to parse with any of the opening classes.

        See `ParsedStatement.consume` for more details.
        """
        for c in cls.opening_classes():
            el = c.consume(statement_iterator, config)
            if el is not None:
                return el
        return None

    @classmethod
    def consume_body(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> ParsedResult[BPST]:
        """Peek into the iterator and try to parse with any of the body classes.

        If the statement cannot be parsed, a UnknownStatement is returned.
        """
        for c in cls.body_classes():
            el = c.consume(statement_iterator, config)
            if el is not None:
                return el
        unkel = next(statement_iterator)
        return UnknownStatement.from_statement(unkel)

    @classmethod
    def consume_closing(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> NullableParsedResult[CPST]:
        """Peek into the iterator and try to parse with any of the opening classes.

        See `ParsedStatement.consume` for more details.
        """
        for c in cls.closing_classes():
            el = c.consume(statement_iterator, config)
            if el is not None:
                return el
        return None

    @classmethod
    def consume_body_closing(
        cls, opening: OPST, statement_iterator: StatementIterator, config: CT
    ) -> Self:
        body: list[ParsedResult[BPST]] = []
        closing: ty.Union[CPST, ParsingError, None] = None
        last_line = opening.end_line
        while closing is None:
            try:
                closing = cls.consume_closing(statement_iterator, config)
                if closing is not None:
                    continue
                el = cls.consume_body(statement_iterator, config)
                body.append(el)
                last_line = el.end_line
            except StopIteration:
                unexpected_end = cls.on_stop_iteration(config)
                unexpected_end.set_position(last_line + 1, 0, last_line + 1, 0)
                return cls(opening, tuple(body), unexpected_end)

        return cls(opening, tuple(body), closing)

    @classmethod
    def consume(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> Union[Self, None]:
        """Try consume the block.

        Possible outcomes:
        1. The opening was not matched, return None.
        2. A subclass of Block, where body and closing migh contain errors.
        """
        opening = cls.consume_opening(statement_iterator, config)
        if opening is None:
            return None

        if isinstance(opening, ParsingError):
            return None

        return cls.consume_body_closing(opening, statement_iterator, config)

    @classmethod
    def on_stop_iteration(cls, config: CT) -> ParsedResult[EOS[CT]]:
        return UnexpectedEOS()


@dataclass(frozen=True)
class BOS(ty.Generic[CT], ParsedStatement[CT]):
    """Beginning of source."""

    # Hasher algorithm name and hexdigest
    content_hash: Hash

    @classmethod
    def from_string_and_config(cls, s: str, config: CT) -> NullableParsedResult[Self]:
        raise RuntimeError("BOS cannot be constructed from_string_and_config")

    @property
    def location(self) -> SourceLocationT:
        return "<undefined>"


@dataclass(frozen=True)
class BOF(ty.Generic[CT], BOS[CT]):
    """Beginning of file."""

    path: pathlib.Path

    # Modification time of the file.
    mtime: float

    @property
    def location(self) -> SourceLocationT:
        return self.path


@dataclass(frozen=True)
class BOR(ty.Generic[CT], BOS[CT]):
    """Beginning of resource."""

    package: str
    resource_name: str

    @property
    def location(self) -> SourceLocationT:
        return self.package, self.resource_name


@dataclass(frozen=True)
class EOS(ty.Generic[CT], ParsedStatement[CT]):
    """End of sequence."""

    @classmethod
    def from_string_and_config(
        cls: type[PST], s: str, config: CT
    ) -> NullableParsedResult[PST]:
        return cls()


class RootBlock(ty.Generic[BPST, CT], Block[BOS[CT], BPST, EOS[CT], CT]):
    """A sequence of statement flanked by the beginning and ending of stream."""

    @classmethod
    def consume_opening(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> NullableParsedResult[BOS[CT]]:
        raise RuntimeError(
            "Implementation error, 'RootBlock.consume_opening' should never be called"
        )

    @classmethod
    def consume(cls, statement_iterator: StatementIterator, config: CT) -> Self:
        block = super().consume(statement_iterator, config)
        if block is None:
            raise RuntimeError(
                "Implementation error, 'RootBlock.consume' should never return None"
            )
        return block

    @classmethod
    def consume_closing(
        cls, statement_iterator: StatementIterator, config: CT
    ) -> NullableParsedResult[EOS[CT]]:
        return None

    @classmethod
    def on_stop_iteration(cls, config: CT) -> ParsedResult[EOS[CT]]:
        return EOS[CT]()


#################
# Source parsing
#################

ResourceT: TypeAlias = tuple[str, str]  # package name, resource name
StrictLocationT: TypeAlias = Union[pathlib.Path, ResourceT]
SourceLocationT: TypeAlias = Union[str, StrictLocationT]


@dataclass(frozen=True)
class ParsedSource(ty.Generic[RBT, CT]):
    parsed_source: RBT

    # Parser configuration.
    config: CT

    @property
    def location(self) -> SourceLocationT:
        if isinstance(self.parsed_source.opening, ParsingError):
            raise self.parsed_source.opening
        return self.parsed_source.opening.location

    @cached_property
    def has_errors(self) -> bool:
        return self.parsed_source.has_errors

    def errors(self) -> ty.Generator[ParsingError, None, None]:
        yield from self.parsed_source.errors


@dataclass(frozen=True)
class CannotParseResourceAsFile(Exception):
    """The requested python package resource cannot be located as a file
    in the file system.
    """

    package: str
    resource_name: str


class Parser(ty.Generic[RBT, CT], GenericInfo):
    """Parser class."""

    #: class to iterate through statements in a source unit.
    _statement_iterator_class: type[StatementIterator] = StatementIterator

    #: Delimiters.
    _delimiters: DelimiterDictT = SPLIT_EOL

    _strip_spaces: bool = True

    #: source file text encoding.
    _encoding: str = "utf-8"

    #: configuration passed to from_string functions.
    _config: CT

    #: try to open resources as files.
    _prefer_resource_as_file: bool

    #: parser algorithm to us. Must be a callable member of hashlib
    _hasher: ty.Callable[
        [
            bytes,
        ],
        HasherProtocol,
    ] = hashlib.blake2b

    def __init__(self, config: CT, prefer_resource_as_file: bool = True):
        self._config = config
        self._prefer_resource_as_file = prefer_resource_as_file

    @classmethod
    def root_boot_class(cls) -> type[RBT]:
        """Class representing the root block class."""
        try:
            return cls.specialization()[RBT]  # type: ignore[misc]
        except KeyError:
            return ty.get_type_hints(cls)["root_boot_class"]  # type: ignore[no-redef]

    def parse(self, source_location: SourceLocationT) -> ParsedSource[RBT, CT]:
        """Parse a file into a ParsedSourceFile or ParsedResource.

        Parameters
        ----------
        source_location:
            if str or pathlib.Path is interpreted as a file.
            if (str, str) is interpreted as (package, resource) using the resource python api.
        """
        if isinstance(source_location, tuple) and len(source_location) == 2:
            if self._prefer_resource_as_file:
                try:
                    return self.parse_resource_from_file(*source_location)
                except CannotParseResourceAsFile:
                    pass
            return self.parse_resource(*source_location)

        if isinstance(source_location, str):
            return self.parse_file(pathlib.Path(source_location))

        if isinstance(source_location, pathlib.Path):
            return self.parse_file(source_location)

        raise TypeError(
            f"Unknown type {type(source_location)}, "
            "use str or pathlib.Path for files or "
            "(package: str, resource_name: str) tuple "
            "for a resource."
        )

    def parse_bytes(
        self, b: bytes, bos: Optional[BOS[CT]] = None
    ) -> ParsedSource[RBT, CT]:
        if bos is None:
            bos = BOS[CT](Hash.from_bytes(self._hasher, b)).set_simple_position(0, 0, 0)

        sic = self._statement_iterator_class(
            b.decode(self._encoding), self._delimiters, self._strip_spaces
        )

        parsed = self.root_boot_class().consume_body_closing(bos, sic, self._config)

        return ParsedSource(
            parsed,
            self._config,
        )

    def parse_file(self, path: pathlib.Path) -> ParsedSource[RBT, CT]:
        """Parse a file into a ParsedSourceFile.

        Parameters
        ----------
        path
            path of the file.
        """
        with path.open(mode="rb") as fi:
            content = fi.read()

        bos = BOF[CT](
            Hash.from_bytes(self._hasher, content), path, path.stat().st_mtime
        ).set_simple_position(0, 0, 0)
        return self.parse_bytes(content, bos)

    def parse_resource_from_file(
        self, package: str, resource_name: str
    ) -> ParsedSource[RBT, CT]:
        """Parse a resource into a ParsedSourceFile, opening as a file.

        Parameters
        ----------
        package
            package name where the resource is located.
        resource_name
            name of the resource
        """
        with resources.as_file(resources.files(package).joinpath(resource_name)) as p:
            path = p.resolve()

        if path.exists():
            return self.parse_file(path)

        raise CannotParseResourceAsFile(package, resource_name)

    def parse_resource(self, package: str, resource_name: str) -> ParsedSource[RBT, CT]:
        """Parse a resource into a ParsedResource.

        Parameters
        ----------
        package
            package name where the resource is located.
        resource_name
            name of the resource
        """
        with resources.files(package).joinpath(resource_name).open("rb") as fi:
            content = fi.read()

        bos = BOR[CT](
            Hash.from_bytes(self._hasher, content), package, resource_name
        ).set_simple_position(0, 0, 0)

        return self.parse_bytes(content, bos)


##########
# Project
##########


class IncludeStatement(ty.Generic[CT], ParsedStatement[CT]):
    """ "Include statements allow to merge files."""

    @property
    def target(self) -> str:
        raise NotImplementedError(
            "IncludeStatement subclasses must implement target property."
        )


class ParsedProject(
    ty.Generic[RBT, CT],
    dict[
        Optional[tuple[StrictLocationT, str]],
        ParsedSource[RBT, CT],
    ],
):
    """Collection of files, independent or connected via IncludeStatement.

    Keys are either an absolute pathname  or a tuple package name, resource name.

    None is the name of the root.

    """

    @cached_property
    def has_errors(self) -> bool:
        return any(el.has_errors for el in self.values())

    def errors(self) -> ty.Generator[ParsingError, None, None]:
        for el in self.values():
            yield from el.errors()

    def _iter_statements(
        self,
        items: ty.Iterable[tuple[Any, Any]],
        seen: set[Any],
        include_only_once: bool,
    ) -> ty.Generator[ParsedStatement[CT], None, None]:
        """Iter all definitions in the order they appear,
        going into the included files.
        """
        for source_location, parsed in items:
            seen.add(source_location)
            for parsed_statement in parsed.parsed_source:
                if isinstance(parsed_statement, IncludeStatement):
                    location = parsed.location, parsed_statement.target
                    if location in seen and include_only_once:
                        raise ValueError(f"{location} was already included.")
                    yield from self._iter_statements(
                        ((location, self[location]),), seen, include_only_once
                    )
                else:
                    yield parsed_statement

    def iter_statements(
        self, include_only_once: bool = True
    ) -> ty.Generator[ParsedStatement[CT], None, None]:
        """Iter all definitions in the order they appear,
        going into the included files.

        Parameters
        ----------
        include_only_once
            if true, each file cannot be included more than once.
        """
        yield from self._iter_statements([(None, self[None])], set(), include_only_once)

    def _iter_blocks(
        self,
        items: ty.Iterable[tuple[Any, Any]],
        seen: set[Any],
        include_only_once: bool,
    ) -> ty.Generator[ParsedStatement[CT], None, None]:
        """Iter all definitions in the order they appear,
        going into the included files.
        """
        for source_location, parsed in items:
            seen.add(source_location)
            for parsed_statement in parsed.parsed_source.iter_blocks():
                if isinstance(parsed_statement, IncludeStatement):
                    location = parsed.location, parsed_statement.target
                    if location in seen and include_only_once:
                        raise ValueError(f"{location} was already included.")
                    yield from self._iter_blocks(
                        ((location, self[location]),), seen, include_only_once
                    )
                else:
                    yield parsed_statement

    def iter_blocks(
        self, include_only_once: bool = True
    ) -> ty.Generator[ParsedStatement[CT], None, None]:
        """Iter all definitions in the order they appear,
        going into the included files.

        Parameters
        ----------
        include_only_once
            if true, each file cannot be included more than once.
        """
        yield from self._iter_blocks([(None, self[None])], set(), include_only_once)


def default_locator(source_location: StrictLocationT, target: str) -> StrictLocationT:
    """Return a new location from current_location and target."""

    if isinstance(source_location, pathlib.Path):
        current_location = pathlib.Path(source_location).resolve()

        if current_location.is_file():
            current_path = current_location.parent
        else:
            current_path = current_location

        target_path = pathlib.Path(target)
        if target_path.is_absolute():
            raise ValueError(
                f"Cannot refer to absolute paths in import statements ({source_location}, {target})."
            )

        tmp = (current_path / target_path).resolve()
        if not tmp.is_relative_to(current_path):
            raise ValueError(
                f"Cannot refer to locations above the current location ({source_location}, {target})"
            )

        return tmp.absolute()

    elif isinstance(source_location, tuple) and len(source_location) == 2:
        return source_location[0], target

    raise TypeError(
        f"Cannot handle type {type(source_location)}, "
        "use str or pathlib.Path for files or "
        "(package: str, resource_name: str) tuple "
        "for a resource."
    )


@no_type_check
def _build_root_block_class_parsed_statement(
    spec: type[ParsedStatement[CT]], config: type[CT]
) -> type[RootBlock[ParsedStatement[CT], CT]]:
    """Build root block class from a single ParsedStatement."""

    @dataclass(frozen=True)
    class CustomRootBlockA(RootBlock[spec, config]):  # type: ignore
        pass

    return CustomRootBlockA


@no_type_check
def _build_root_block_class_block(
    spec: type[Block[OPST, BPST, CPST, CT]],
    config: type[CT],
) -> type[RootBlock[Block[OPST, BPST, CPST, CT], CT]]:
    """Build root block class from a single ParsedStatement."""

    @dataclass(frozen=True)
    class CustomRootBlockA(RootBlock[spec, config]):  # type: ignore
        pass

    return CustomRootBlockA


@no_type_check
def _build_root_block_class_parsed_statement_it(
    spec: tuple[type[Union[ParsedStatement[CT], Block[OPST, BPST, CPST, CT]]]],
    config: type[CT],
) -> type[RootBlock[ParsedStatement[CT], CT]]:
    """Build root block class from iterable ParsedStatement."""

    @dataclass(frozen=True)
    class CustomRootBlockA(RootBlock[Union[spec], config]):  # type: ignore
        pass

    return CustomRootBlockA


@no_type_check
def _build_parser_class_root_block(
    spec: type[RootBlock[BPST, CT]],
    *,
    strip_spaces: bool = True,
    delimiters: Optional[DelimiterDictT] = None,
) -> type[Parser[RootBlock[BPST, CT], CT]]:
    class CustomParser(Parser[spec, spec.specialization()[CT]]):  # type: ignore
        _delimiters: DelimiterDictT = delimiters or SPLIT_EOL
        _strip_spaces: bool = strip_spaces

    return CustomParser


@no_type_check
def build_parser_class(
    spec: Union[
        type[
            Union[
                Parser[RBT, CT],
                RootBlock[BPST, CT],
                Block[OPST, BPST, CPST, CT],
                ParsedStatement[CT],
            ]
        ],
        ty.Iterable[type[ParsedStatement[CT]]],
    ],
    config: CT = None,
    strip_spaces: bool = True,
    delimiters: Optional[DelimiterDictT] = None,
) -> type[
    Union[
        Parser[RBT, CT],
        Parser[RootBlock[BPST, CT], CT],
        Parser[RootBlock[Block[OPST, BPST, CPST, CT], CT], CT],
    ]
]:
    """Build a custom parser class.

    Parameters
    ----------
    spec
        RootBlock derived class.
    strip_spaces : bool
        if True, spaces will be stripped for each statement before calling
        ``from_string_and_config``.
    delimiters : dict
        Specify how the source file is split into statements (See below).

    Delimiters dictionary
    ---------------------
        The delimiters are specified with the keys of the delimiters dict.
    The dict files can be used to further customize the iterator. Each
    consist of a tuple of two elements:
      1. A value of the DelimiterMode to indicate what to do with the
         delimiter string: skip it, attach keep it with previous or next string
      2. A boolean indicating if parsing should stop after fiSBT
         encountering this delimiter.
    """

    if isinstance(spec, type):
        if issubclass(spec, Parser):
            CustomParser = spec

        elif issubclass(spec, RootBlock):
            CustomParser = _build_parser_class_root_block(
                spec, strip_spaces=strip_spaces, delimiters=delimiters
            )

        elif issubclass(spec, Block):
            CustomRootBlock = _build_root_block_class_block(spec, config.__class__)
            CustomParser = _build_parser_class_root_block(
                CustomRootBlock, strip_spaces=strip_spaces, delimiters=delimiters
            )

        elif issubclass(spec, ParsedStatement):
            CustomRootBlock = _build_root_block_class_parsed_statement(
                spec, config.__class__
            )
            CustomParser = _build_parser_class_root_block(
                CustomRootBlock, strip_spaces=strip_spaces, delimiters=delimiters
            )

        else:
            raise TypeError(
                "`spec` must be of type Parser, Block, RootBlock or tuple of type Block or ParsedStatement, "
                f"not {type(spec)}"
            )

    elif isinstance(spec, (tuple, list)):
        CustomRootBlock = _build_root_block_class_parsed_statement_it(
            spec, config.__class__
        )
        CustomParser = _build_parser_class_root_block(
            CustomRootBlock, strip_spaces=strip_spaces, delimiters=delimiters
        )

    else:
        raise

    return CustomParser


@no_type_check
def parse(
    entry_point: SourceLocationT,
    spec: Union[
        type[
            Union[
                Parser[RBT, CT],
                RootBlock[BPST, CT],
                Block[OPST, BPST, CPST, CT],
                ParsedStatement[CT],
            ]
        ],
        ty.Iterable[type[ParsedStatement[CT]]],
    ],
    config: CT = None,
    *,
    strip_spaces: bool = True,
    delimiters: Optional[DelimiterDictT] = None,
    locator: ty.Callable[[SourceLocationT, str], StrictLocationT] = default_locator,
    prefer_resource_as_file: bool = True,
    **extra_parser_kwargs: Any,
) -> Union[ParsedProject[RBT, CT], ParsedProject[RootBlock[BPST, CT], CT]]:
    """Parse sources into a ParsedProject dictionary.

    Parameters
    ----------
    entry_point
        file or resource, given as (package_name, resource_name).
    spec
        specification of the content to parse. Can be one of the following things:
        - Parser class.
        - Block or ParsedStatement derived class.
        - ty.Iterable of Block or ParsedStatement derived class.
        - RootBlock derived class.
    config
        a configuration object that will be passed to `from_string_and_config`
        classmethod.
    strip_spaces : bool
        if True, spaces will be stripped for each statement before calling
        ``from_string_and_config``.
    delimiters : dict
        Specify how the source file is split into statements (See below).
    locator : Callable
        function that takes the current location and a target of an IncludeStatement
        and returns a new location.
    prefer_resource_as_file : bool
        if True, resources will try to be located in the filesystem if
        available.
    extra_parser_kwargs
        extra keyword arguments to be given to the parser.

    Delimiters dictionary
    ---------------------
        The delimiters are specified with the keys of the delimiters dict.
    The dict files can be used to further customize the iterator. Each
    consist of a tuple of two elements:
      1. A value of the DelimiterMode to indicate what to do with the
         delimiter string: skip it, attach keep it with previous or next string
      2. A boolean indicating if parsing should stop after fiSBT
         encountering this delimiter.
    """

    CustomParser = build_parser_class(spec, config, strip_spaces, delimiters)
    parser = CustomParser(
        config, prefer_resource_as_file=prefer_resource_as_file, **extra_parser_kwargs
    )

    pp = ParsedProject()

    pending: list[tuple[SourceLocationT, str]] = []
    if isinstance(entry_point, (str, pathlib.Path)):
        entry_point = pathlib.Path(entry_point)
        if not entry_point.is_absolute():
            entry_point = pathlib.Path.cwd() / entry_point

    elif not (isinstance(entry_point, tuple) and len(entry_point) == 2):
        raise TypeError(
            f"Cannot handle type {type(entry_point)}, "
            "use str or pathlib.Path for files or "
            "(package: str, resource_name: str) tuple "
            "for a resource."
        )

    pp[None] = parsed = parser.parse(entry_point)
    pending.extend(
        (parsed.location, el.target)
        for el in parsed.parsed_source.filter_by(IncludeStatement)
    )

    while pending:
        source_location, target = pending.pop(0)
        pp[(source_location, target)] = parsed = parser.parse(
            locator(source_location, target)
        )
        pending.extend(
            (parsed.location, el.target)
            for el in parsed.parsed_source.filter_by(IncludeStatement)
        )

    return pp


@no_type_check
def parse_bytes(
    content: bytes,
    spec: Union[
        type[
            Union[
                Parser[RBT, CT],
                RootBlock[BPST, CT],
                Block[OPST, BPST, CPST, CT],
                ParsedStatement[CT],
            ]
        ],
        ty.Iterable[type[ParsedStatement[CT]]],
    ],
    config: Optional[CT] = None,
    *,
    strip_spaces: bool,
    delimiters: Optional[DelimiterDictT],
    **extra_parser_kwargs: Any,
) -> ParsedProject[
    Union[RBT, RootBlock[BPST, CT], RootBlock[ParsedStatement[CT], CT]], CT
]:
    """Parse sources into a ParsedProject dictionary.

    Parameters
    ----------
    content
        bytes.
    spec
        specification of the content to parse. Can be one of the following things:
        - Parser class.
        - Block or ParsedStatement derived class.
        - ty.Iterable of Block or ParsedStatement derived class.
        - RootBlock derived class.
    config
        a configuration object that will be passed to `from_string_and_config`
        classmethod.
    strip_spaces : bool
        if True, spaces will be stripped for each statement before calling
        ``from_string_and_config``.
    delimiters : dict
        Specify how the source file is split into statements (See below).
    """

    CustomParser = build_parser_class(spec, config, strip_spaces, delimiters)

    parser = CustomParser(config, prefer_resource_as_file=False, **extra_parser_kwargs)

    pp = ParsedProject()

    pp[None] = parsed = parser.parse_bytes(content)

    if any(parsed.parsed_source.filter_by(IncludeStatement)):
        raise ValueError("parse_bytes does not support using an IncludeStatement")

    return pp



"""
    pint.facets.formatting.registry
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2022 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from typing import Generic, Any

from ...compat import TypeAlias
from ..plain import GenericPlainRegistry, QuantityT, UnitT
from . import objects


class GenericFormattingRegistry(
    Generic[QuantityT, UnitT], GenericPlainRegistry[QuantityT, UnitT]
):
    pass


class FormattingRegistry(
    GenericFormattingRegistry[objects.FormattingQuantity[Any], objects.FormattingUnit]
):
    Quantity: TypeAlias = objects.FormattingQuantity[Any]
    Unit: TypeAlias = objects.FormattingUnit



"""
    pint.delegates.txt_defparser
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parser for the original textual Pint Definition file.

    :copyright: 2022 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .defparser import DefParser

__all__ = [
    "DefParser",
]



"""
    pint.facets.formatting.objects
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2022 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import re
from typing import Any, Generic

from ...compat import babel_parse, ndarray, np
from ...formatting import (
    _pretty_fmt_exponent,
    extract_custom_flags,
    format_unit,
    ndarray_to_latex,
    remove_custom_flags,
    siunitx_format_unit,
    split_format,
)
from ...util import UnitsContainer, iterable

from ..plain import PlainQuantity, PlainUnit, MagnitudeT


class FormattingQuantity(Generic[MagnitudeT], PlainQuantity[MagnitudeT]):
    _exp_pattern = re.compile(r"([0-9]\.?[0-9]*)e(-?)\+?0*([0-9]+)")

    def __format__(self, spec: str) -> str:
        if self._REGISTRY.fmt_locale is not None:
            return self.format_babel(spec)

        mspec, uspec = split_format(
            spec, self.default_format, self._REGISTRY.separate_format_defaults
        )

        # If Compact is selected, do it at the beginning
        if "#" in spec:
            # TODO: don't replace '#'
            mspec = mspec.replace("#", "")
            uspec = uspec.replace("#", "")
            obj = self.to_compact()
        else:
            obj = self

        if "L" in uspec:
            allf = plain_allf = r"{}\ {}"
        elif "H" in uspec:
            allf = plain_allf = "{} {}"
            if iterable(obj.magnitude):
                # Use HTML table instead of plain text template for array-likes
                allf = (
                    "<table><tbody>"
                    "<tr><th>Magnitude</th>"
                    "<td style='text-align:left;'>{}</td></tr>"
                    "<tr><th>Units</th><td style='text-align:left;'>{}</td></tr>"
                    "</tbody></table>"
                )
        else:
            allf = plain_allf = "{} {}"

        if "Lx" in uspec:
            # the LaTeX siunitx code
            # TODO: add support for extracting options
            opts = ""
            ustr = siunitx_format_unit(obj.units._units, obj._REGISTRY)
            allf = r"\SI[%s]{{{}}}{{{}}}" % opts
        else:
            # Hand off to unit formatting
            # TODO: only use `uspec` after completing the deprecation cycle
            ustr = format(obj.units, mspec + uspec)

        # mspec = remove_custom_flags(spec)
        if "H" in uspec:
            # HTML formatting
            if hasattr(obj.magnitude, "_repr_html_"):
                # If magnitude has an HTML repr, nest it within Pint's
                mstr = obj.magnitude._repr_html_()
            else:
                if isinstance(self.magnitude, ndarray):
                    # Use custom ndarray text formatting with monospace font
                    formatter = f"{{:{mspec}}}"
                    # Need to override for scalars, which are detected as iterable,
                    # and don't respond to printoptions.
                    if self.magnitude.ndim == 0:
                        allf = plain_allf = "{} {}"
                        mstr = formatter.format(obj.magnitude)
                    else:
                        with np.printoptions(
                            formatter={"float_kind": formatter.format}
                        ):
                            mstr = (
                                "<pre>"
                                + format(obj.magnitude).replace("\n", "<br>")
                                + "</pre>"
                            )
                elif not iterable(obj.magnitude):
                    # Use plain text for scalars
                    mstr = format(obj.magnitude, mspec)
                else:
                    # Use monospace font for other array-likes
                    mstr = (
                        "<pre>"
                        + format(obj.magnitude, mspec).replace("\n", "<br>")
                        + "</pre>"
                    )
        elif isinstance(self.magnitude, ndarray):
            if "L" in uspec:
                # Use ndarray LaTeX special formatting
                mstr = ndarray_to_latex(obj.magnitude, mspec)
            else:
                # Use custom ndarray text formatting--need to handle scalars differently
                # since they don't respond to printoptions
                formatter = f"{{:{mspec}}}"
                if obj.magnitude.ndim == 0:
                    mstr = formatter.format(obj.magnitude)
                else:
                    with np.printoptions(formatter={"float_kind": formatter.format}):
                        mstr = format(obj.magnitude).replace("\n", "")
        else:
            mstr = format(obj.magnitude, mspec).replace("\n", "")

        if "L" in uspec and "Lx" not in uspec:
            mstr = self._exp_pattern.sub(r"\1\\times 10^{\2\3}", mstr)
        elif "H" in uspec or "P" in uspec:
            m = self._exp_pattern.match(mstr)
            _exp_formatter = (
                _pretty_fmt_exponent if "P" in uspec else lambda s: f"<sup>{s}</sup>"
            )
            if m:
                exp = int(m.group(2) + m.group(3))
                mstr = self._exp_pattern.sub(r"\1×10" + _exp_formatter(exp), mstr)

        if allf == plain_allf and ustr.startswith("1 /"):
            # Write e.g. "3 / s" instead of "3 1 / s"
            ustr = ustr[2:]
        return allf.format(mstr, ustr).strip()

    def _repr_pretty_(self, p, cycle):
        if cycle:
            super()._repr_pretty_(p, cycle)
        else:
            p.pretty(self.magnitude)
            p.text(" ")
            p.pretty(self.units)

    def format_babel(self, spec: str = "", **kwspec: Any) -> str:
        spec = spec or self.default_format

        # standard cases
        if "#" in spec:
            spec = spec.replace("#", "")
            obj = self.to_compact()
        else:
            obj = self
        kwspec = kwspec.copy()
        if "length" in kwspec:
            kwspec["babel_length"] = kwspec.pop("length")

        loc = kwspec.get("locale", self._REGISTRY.fmt_locale)
        if loc is None:
            raise ValueError("Provide a `locale` value to localize translation.")

        kwspec["locale"] = babel_parse(loc)
        kwspec["babel_plural_form"] = kwspec["locale"].plural_form(obj.magnitude)
        return "{} {}".format(
            format(obj.magnitude, remove_custom_flags(spec)),
            obj.units.format_babel(spec, **kwspec),
        ).replace("\n", "")

    def __str__(self) -> str:
        if self._REGISTRY.fmt_locale is not None:
            return self.format_babel()

        return format(self)


class FormattingUnit(PlainUnit):
    def __str__(self):
        return format(self)

    def __format__(self, spec) -> str:
        _, uspec = split_format(
            spec, self.default_format, self._REGISTRY.separate_format_defaults
        )
        if "~" in uspec:
            if not self._units:
                return ""
            units = UnitsContainer(
                {
                    self._REGISTRY._get_symbol(key): value
                    for key, value in self._units.items()
                }
            )
            uspec = uspec.replace("~", "")
        else:
            units = self._units

        return format_unit(units, uspec, registry=self._REGISTRY)

    def format_babel(self, spec="", locale=None, **kwspec: Any) -> str:
        spec = spec or extract_custom_flags(self.default_format)

        if "~" in spec:
            if self.dimensionless:
                return ""
            units = UnitsContainer(
                {
                    self._REGISTRY._get_symbol(key): value
                    for key, value in self._units.items()
                }
            )
            spec = spec.replace("~", "")
        else:
            units = self._units

        locale = self._REGISTRY.fmt_locale if locale is None else locale

        if locale is None:
            raise ValueError("Provide a `locale` value to localize translation.")
        else:
            kwspec["locale"] = babel_parse(locale)

        return units.format_babel(spec, registry=self._REGISTRY, **kwspec)



"""
    pint.facets.context
    ~~~~~~~~~~~~~~~~~~~

    Adds pint the capability to contexts: predefined conversions
    between incompatible dimensions.

    :copyright: 2022 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from .definitions import ContextDefinition
from .objects import Context
from .registry import ContextRegistry, GenericContextRegistry

__all__ = ["ContextDefinition", "Context", "ContextRegistry", "GenericContextRegistry"]



"""
    pint.util
    ~~~~~~~~~

    Miscellaneous functions for pint.

    :copyright: 2016 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import logging
import math
import operator
import re
from collections.abc import Mapping, Iterable, Iterator
from fractions import Fraction
from functools import lru_cache, partial
from logging import NullHandler
from numbers import Number
from token import NAME, NUMBER
import tokenize
import types
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Callable,
    TypeVar,
    Any,
    Optional,
)
from collections.abc import Hashable, Generator

from .compat import NUMERIC_TYPES, Self
from .errors import DefinitionSyntaxError
from .formatting import format_unit
from .pint_eval import build_eval_tree
from . import pint_eval

from ._typing import Scalar

if TYPE_CHECKING:
    from ._typing import QuantityOrUnitLike
    from .registry import UnitRegistry


logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())

T = TypeVar("T")
TH = TypeVar("TH", bound=Hashable)
TT = TypeVar("TT", bound=type)

# TODO: Change when Python 3.10 becomes minimal version.
# ItMatrix: TypeAlias = Iterable[Iterable[PintScalar]]
# Matrix: TypeAlias = list[list[PintScalar]]
ItMatrix = Iterable[Iterable[Scalar]]
Matrix = list[list[Scalar]]


def _noop(x: T) -> T:
    return x


def matrix_to_string(
    matrix: ItMatrix,
    row_headers: Optional[Iterable[str]] = None,
    col_headers: Optional[Iterable[str]] = None,
    fmtfun: Callable[
        [
            Scalar,
        ],
        str,
    ] = "{:0.0f}".format,
) -> str:
    """Return a string representation of a matrix.

    Parameters
    ----------
    matrix
        A matrix given as an iterable of an iterable of numbers.
    row_headers
        An iterable of strings to serve as row headers.
        (default = None, meaning no row headers are printed.)
    col_headers
        An iterable of strings to serve as column headers.
        (default = None, meaning no col headers are printed.)
    fmtfun
        A callable to convert a number into string.
        (default = `"{:0.0f}".format`)

    Returns
    -------
    str
        String representation of the matrix.
    """
    ret: list[str] = []
    if col_headers:
        ret.append(("\t" if row_headers else "") + "\t".join(col_headers))
    if row_headers:
        ret += [
            rh + "\t" + "\t".join(fmtfun(f) for f in row)
            for rh, row in zip(row_headers, matrix)
        ]
    else:
        ret += ["\t".join(fmtfun(f) for f in row) for row in matrix]

    return "\n".join(ret)


def transpose(matrix: ItMatrix) -> Matrix:
    """Return the transposed version of a matrix.

    Parameters
    ----------
    matrix
        A matrix given as an iterable of an iterable of numbers.

    Returns
    -------
    Matrix
        The transposed version of the matrix.
    """
    return [list(val) for val in zip(*matrix)]


def matrix_apply(
    matrix: ItMatrix,
    func: Callable[
        [
            Scalar,
        ],
        Scalar,
    ],
) -> Matrix:
    """Apply a function to individual elements within a matrix.

    Parameters
    ----------
    matrix
        A matrix given as an iterable of an iterable of numbers.
    func
        A callable that converts a number to another.

    Returns
    -------
    A new matrix in which each element has been replaced by new one.
    """
    return [[func(x) for x in row] for row in matrix]


def column_echelon_form(
    matrix: ItMatrix, ntype: type = Fraction, transpose_result: bool = False
) -> tuple[Matrix, Matrix, list[int]]:
    """Calculate the column echelon form using Gaussian elimination.

    Parameters
    ----------
    matrix
        A 2D matrix as nested list.
    ntype
        The numerical type to use in the calculation.
        (default = Fraction)
    transpose_result
        Indicates if the returned matrix should be transposed.
        (default = False)

    Returns
    -------
    ech_matrix
        Column echelon form.
    id_matrix
        Transformed identity matrix.
    swapped
        Swapped rows.
    """

    _transpose: Callable[
        [
            ItMatrix,
        ],
        Matrix,
    ] = (
        transpose if transpose_result else _noop
    )

    ech_matrix = matrix_apply(
        transpose(matrix),
        lambda x: ntype.from_float(x) if isinstance(x, float) else ntype(x),  # type: ignore
    )

    rows, cols = len(ech_matrix), len(ech_matrix[0])
    # M = [[ntype(x) for x in row] for row in M]
    id_matrix: list[list[Scalar]] = [  # noqa: E741
        [ntype(1) if n == nc else ntype(0) for nc in range(rows)] for n in range(rows)
    ]

    swapped: list[int] = []
    lead = 0
    for r in range(rows):
        if lead >= cols:
            return _transpose(ech_matrix), _transpose(id_matrix), swapped
        s = r
        while ech_matrix[s][lead] == 0:  # type: ignore
            s += 1
            if s != rows:
                continue
            s = r
            lead += 1
            if cols == lead:
                return _transpose(ech_matrix), _transpose(id_matrix), swapped

        ech_matrix[s], ech_matrix[r] = ech_matrix[r], ech_matrix[s]
        id_matrix[s], id_matrix[r] = id_matrix[r], id_matrix[s]

        swapped.append(s)
        lv = ech_matrix[r][lead]
        ech_matrix[r] = [mrx / lv for mrx in ech_matrix[r]]
        id_matrix[r] = [mrx / lv for mrx in id_matrix[r]]

        for s in range(rows):
            if s == r:
                continue
            lv = ech_matrix[s][lead]
            ech_matrix[s] = [
                iv - lv * rv for rv, iv in zip(ech_matrix[r], ech_matrix[s])
            ]
            id_matrix[s] = [iv - lv * rv for rv, iv in zip(id_matrix[r], id_matrix[s])]

        lead += 1

    return _transpose(ech_matrix), _transpose(id_matrix), swapped


def pi_theorem(quantities: dict[str, Any], registry: Optional[UnitRegistry] = None):
    """Builds dimensionless quantities using the Buckingham π theorem

    Parameters
    ----------
    quantities : dict
        mapping between variable name and units
    registry :
         (default value = None)

    Returns
    -------
    type
        a list of dimensionless quantities expressed as dicts

    """

    # Preprocess input and build the dimensionality Matrix
    quant = []
    dimensions = set()

    if registry is None:
        getdim = _noop
        non_int_type = float
    else:
        getdim = registry.get_dimensionality
        non_int_type = registry.non_int_type

    for name, value in quantities.items():
        if isinstance(value, str):
            value = ParserHelper.from_string(value, non_int_type=non_int_type)
        if isinstance(value, dict):
            dims = getdim(registry.UnitsContainer(value))
        elif not hasattr(value, "dimensionality"):
            dims = getdim(value)
        else:
            dims = value.dimensionality

        if not registry and any(not key.startswith("[") for key in dims):
            logger.warning(
                "A non dimension was found and a registry was not provided. "
                "Assuming that it is a dimension name: {}.".format(dims)
            )

        quant.append((name, dims))
        dimensions = dimensions.union(dims.keys())

    dimensions = list(dimensions)

    # Calculate dimensionless  quantities
    matrix = [
        [dimensionality[dimension] for name, dimensionality in quant]
        for dimension in dimensions
    ]

    ech_matrix, id_matrix, pivot = column_echelon_form(matrix, transpose_result=False)

    # Collect results
    # Make all numbers integers and minimize the number of negative exponents.
    # Remove zeros
    results = []
    for rowm, rowi in zip(ech_matrix, id_matrix):
        if any(el != 0 for el in rowm):
            continue
        max_den = max(f.denominator for f in rowi)
        neg = -1 if sum(f < 0 for f in rowi) > sum(f > 0 for f in rowi) else 1
        results.append(
            {
                q[0]: neg * f.numerator * max_den / f.denominator
                for q, f in zip(quant, rowi)
                if f.numerator != 0
            }
        )
    return results


def solve_dependencies(
    dependencies: dict[TH, set[TH]]
) -> Generator[set[TH], None, None]:
    """Solve a dependency graph.

    Parameters
    ----------
    dependencies :
        dependency dictionary. For each key, the value is an iterable indicating its
        dependencies.

    Yields
    ------
    set
        iterator of sets, each containing keys of independents tasks dependent only of
        the previous tasks in the list.

    Raises
    ------
    ValueError
        if a cyclic dependency is found.
    """
    while dependencies:
        # values not in keys (items without dep)
        t = {i for v in dependencies.values() for i in v} - dependencies.keys()
        # and keys without value (items without dep)
        t.update(k for k, v in dependencies.items() if not v)
        # can be done right away
        if not t:
            raise ValueError(
                "Cyclic dependencies exist among these items: {}".format(
                    ", ".join(repr(x) for x in dependencies.items())
                )
            )
        # and cleaned up
        dependencies = {k: v - t for k, v in dependencies.items() if v}
        yield t


def find_shortest_path(
    graph: dict[TH, set[TH]], start: TH, end: TH, path: Optional[list[TH]] = None
):
    """Find shortest path between two nodes within a graph.

    Parameters
    ----------
    graph
        A graph given as a mapping of nodes
        to a set of all connected nodes to it.
    start
        Starting node.
    end
        End node.
    path
        Path to prepend to the one found.
        (default = None, empty path.)

    Returns
    -------
    list[TH]
        The shortest path between two nodes.
    """
    path = (path or []) + [start]
    if start == end:
        return path

    # TODO: raise ValueError when start not in graph
    if start not in graph:
        return None

    shortest = None
    for node in graph[start]:
        if node not in path:
            newpath = find_shortest_path(graph, node, end, path)
            if newpath:
                if not shortest or len(newpath) < len(shortest):
                    shortest = newpath

    return shortest


def find_connected_nodes(
    graph: dict[TH, set[TH]], start: TH, visited: Optional[set[TH]] = None
) -> Optional[set[TH]]:
    """Find all nodes connected to a start node within a graph.

    Parameters
    ----------
    graph
        A graph given as a mapping of nodes
        to a set of all connected nodes to it.
    start
        Starting node.
    visited
        Mutable set to collect visited nodes.
        (default = None, empty set)

    Returns
    -------
    set[TH]
        The shortest path between two nodes.
    """

    # TODO: raise ValueError when start not in graph
    if start not in graph:
        return None

    visited = visited or set()
    visited.add(start)

    for node in graph[start]:
        if node not in visited:
            find_connected_nodes(graph, node, visited)

    return visited


class udict(dict[str, Scalar]):
    """Custom dict implementing __missing__."""

    def __missing__(self, key: str):
        return 0

    def copy(self: Self) -> Self:
        return udict(self)


class UnitsContainer(Mapping[str, Scalar]):
    """The UnitsContainer stores the product of units and their respective
    exponent and implements the corresponding operations.

    UnitsContainer is a read-only mapping. All operations (even in place ones)
    return new instances.

    Parameters
    ----------
    non_int_type
        Numerical type used for non integer values.
    """

    __slots__ = ("_d", "_hash", "_one", "_non_int_type")

    _d: udict
    _hash: Optional[int]
    _one: Scalar
    _non_int_type: type

    def __init__(
        self, *args: Any, non_int_type: Optional[type] = None, **kwargs: Any
    ) -> None:
        if args and isinstance(args[0], UnitsContainer):
            default_non_int_type = args[0]._non_int_type
        else:
            default_non_int_type = float

        self._non_int_type = non_int_type or default_non_int_type

        if self._non_int_type is float:
            self._one = 1
        else:
            self._one = self._non_int_type("1")

        d = udict(*args, **kwargs)
        self._d = d
        for key, value in d.items():
            if not isinstance(key, str):
                raise TypeError(f"key must be a str, not {type(key)}")
            if not isinstance(value, Number):
                raise TypeError(f"value must be a number, not {type(value)}")
            if not isinstance(value, int) and not isinstance(value, self._non_int_type):
                d[key] = self._non_int_type(value)
        self._hash = None

    def copy(self: Self) -> Self:
        """Create a copy of this UnitsContainer."""
        return self.__copy__()

    def add(self: Self, key: str, value: Number) -> Self:
        """Create a new UnitsContainer adding value to
        the value existing for a given key.

        Parameters
        ----------
        key
            unit to which the value will be added.
        value
            value to be added.

        Returns
        -------
        UnitsContainer
            A copy of this container.
        """
        newval = self._d[key] + value
        new = self.copy()
        if newval:
            new._d[key] = newval
        else:
            new._d.pop(key)
        new._hash = None
        return new

    def remove(self: Self, keys: Iterable[str]) -> Self:
        """Create a new UnitsContainer purged from given entries.

        Parameters
        ----------
        keys
            Iterable of keys (units) to remove.

        Returns
        -------
        UnitsContainer
            A copy of this container.
        """
        new = self.copy()
        for k in keys:
            new._d.pop(k)
        new._hash = None
        return new

    def rename(self: Self, oldkey: str, newkey: str) -> Self:
        """Create a new UnitsContainer in which an entry has been renamed.

        Parameters
        ----------
        oldkey
            Existing key (unit).
        newkey
            New key (unit).

        Returns
        -------
        UnitsContainer
            A copy of this container.
        """
        new = self.copy()
        new._d[newkey] = new._d.pop(oldkey)
        new._hash = None
        return new

    def __iter__(self) -> Iterator[str]:
        return iter(self._d)

    def __len__(self) -> int:
        return len(self._d)

    def __getitem__(self, key: str) -> Scalar:
        return self._d[key]

    def __contains__(self, key: str) -> bool:
        return key in self._d

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = hash(frozenset(self._d.items()))
        return self._hash

    # Only needed by pickle protocol 0 and 1 (used by pytables)
    def __getstate__(self) -> tuple[udict, Scalar, type]:
        return self._d, self._one, self._non_int_type

    def __setstate__(self, state: tuple[udict, Scalar, type]):
        self._d, self._one, self._non_int_type = state
        self._hash = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, UnitsContainer):
            # UnitsContainer.__hash__(self) is not the same as hash(self); see
            # ParserHelper.__hash__ and __eq__.
            # Different hashes guarantee that the actual contents are different, but
            # identical hashes give no guarantee of equality.
            # e.g. in CPython, hash(-1) == hash(-2)
            if UnitsContainer.__hash__(self) != UnitsContainer.__hash__(other):
                return False
            other = other._d

        elif isinstance(other, str):
            try:
                other = ParserHelper.from_string(other, self._non_int_type)
            except DefinitionSyntaxError:
                return False

            other = other._d

        return dict.__eq__(self._d, other)

    def __str__(self) -> str:
        return self.__format__("")

    def __repr__(self) -> str:
        tmp = "{%s}" % ", ".join(
            [f"'{key}': {value}" for key, value in sorted(self._d.items())]
        )
        return f"<UnitsContainer({tmp})>"

    def __format__(self, spec: str) -> str:
        return format_unit(self, spec)

    def format_babel(self, spec: str, registry=None, **kwspec) -> str:
        return format_unit(self, spec, registry=registry, **kwspec)

    def __copy__(self):
        # Skip expensive health checks performed by __init__
        out = object.__new__(self.__class__)
        out._d = self._d.copy()
        out._hash = self._hash
        out._non_int_type = self._non_int_type
        out._one = self._one
        return out

    def __mul__(self, other: Any):
        if not isinstance(other, self.__class__):
            err = "Cannot multiply UnitsContainer by {}"
            raise TypeError(err.format(type(other)))

        new = self.copy()
        for key, value in other.items():
            new._d[key] += value
            if new._d[key] == 0:
                del new._d[key]

        new._hash = None
        return new

    __rmul__ = __mul__

    def __pow__(self, other: Any):
        if not isinstance(other, NUMERIC_TYPES):
            err = "Cannot power UnitsContainer by {}"
            raise TypeError(err.format(type(other)))

        new = self.copy()
        for key, value in new._d.items():
            new._d[key] *= other
        new._hash = None
        return new

    def __truediv__(self, other: Any):
        if not isinstance(other, self.__class__):
            err = "Cannot divide UnitsContainer by {}"
            raise TypeError(err.format(type(other)))

        new = self.copy()
        for key, value in other.items():
            new._d[key] -= value
            if new._d[key] == 0:
                del new._d[key]

        new._hash = None
        return new

    def __rtruediv__(self, other: Any):
        if not isinstance(other, self.__class__) and other != 1:
            err = "Cannot divide {} by UnitsContainer"
            raise TypeError(err.format(type(other)))

        return self**-1


class ParserHelper(UnitsContainer):
    """The ParserHelper stores in place the product of variables and
    their respective exponent and implements the corresponding operations.
    It also provides a scaling factor.

    For example:
        `3 * m ** 2` becomes ParserHelper(3, m=2)

    Briefly is a UnitsContainer with a scaling factor.

    ParserHelper is a read-only mapping. All operations (even in place ones)
    return new instances.

    WARNING : The hash value used does not take into account the scale
    attribute so be careful if you use it as a dict key and then two unequal
    object can have the same hash.

    Parameters
    ----------
    scale
        Scaling factor.
        (default = 1)
    **kwargs
        Used to populate the dict of units and exponents.
    """

    __slots__ = ("scale",)

    scale: Scalar

    def __init__(self, scale: Scalar = 1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale = scale

    @classmethod
    def from_word(cls, input_word: str, non_int_type: type = float) -> ParserHelper:
        """Creates a ParserHelper object with a single variable with exponent one.

        Equivalent to: ParserHelper(1, {input_word: 1})

        Parameters
        ----------
        input_word

        non_int_type
            Numerical type used for non integer values.
        """
        if non_int_type is float:
            return cls(1, [(input_word, 1)], non_int_type=non_int_type)
        else:
            ONE = non_int_type("1")
            return cls(ONE, [(input_word, ONE)], non_int_type=non_int_type)

    @classmethod
    def eval_token(cls, token: tokenize.TokenInfo, non_int_type: type = float):
        token_type = token.type
        token_text = token.string
        if token_type == NUMBER:
            if non_int_type is float:
                try:
                    return int(token_text)
                except ValueError:
                    return float(token_text)
            else:
                return non_int_type(token_text)
        elif token_type == NAME:
            return ParserHelper.from_word(token_text, non_int_type=non_int_type)
        else:
            raise Exception("unknown token type")

    @classmethod
    @lru_cache
    def from_string(cls, input_string: str, non_int_type: type = float) -> ParserHelper:
        """Parse linear expression mathematical units and return a quantity object.

        Parameters
        ----------
        input_string

        non_int_type
            Numerical type used for non integer values.
        """
        if not input_string:
            return cls(non_int_type=non_int_type)

        input_string = string_preprocessor(input_string)
        if "[" in input_string:
            input_string = input_string.replace("[", "__obra__").replace(
                "]", "__cbra__"
            )
            reps = True
        else:
            reps = False

        gen = pint_eval.tokenizer(input_string)
        ret = build_eval_tree(gen).evaluate(
            partial(cls.eval_token, non_int_type=non_int_type)
        )

        if isinstance(ret, Number):
            return cls(ret, non_int_type=non_int_type)

        if reps:
            ret = cls(
                ret.scale,
                {
                    key.replace("__obra__", "[").replace("__cbra__", "]"): value
                    for key, value in ret.items()
                },
                non_int_type=non_int_type,
            )

        for k in list(ret):
            if k.lower() == "nan":
                del ret._d[k]
                ret.scale = non_int_type(math.nan)

        return ret

    def __copy__(self):
        new = super().__copy__()
        new.scale = self.scale
        return new

    def copy(self):
        return self.__copy__()

    def __hash__(self):
        if self.scale != 1:
            mess = "Only scale 1 ParserHelper instance should be considered hashable"
            raise ValueError(mess)
        return super().__hash__()

    # Only needed by pickle protocol 0 and 1 (used by pytables)
    def __getstate__(self):
        return super().__getstate__() + (self.scale,)

    def __setstate__(self, state):
        super().__setstate__(state[:-1])
        self.scale = state[-1]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ParserHelper):
            return self.scale == other.scale and super().__eq__(other)
        elif isinstance(other, str):
            return self == ParserHelper.from_string(other, self._non_int_type)
        elif isinstance(other, Number):
            return self.scale == other and not len(self._d)

        return self.scale == 1 and super().__eq__(other)

    def operate(self, items, op=operator.iadd, cleanup: bool = True):
        d = udict(self._d)
        for key, value in items:
            d[key] = op(d[key], value)

        if cleanup:
            keys = [key for key, value in d.items() if value == 0]
            for key in keys:
                del d[key]

        return self.__class__(self.scale, d, non_int_type=self._non_int_type)

    def __str__(self):
        tmp = "{%s}" % ", ".join(
            [f"'{key}': {value}" for key, value in sorted(self._d.items())]
        )
        return f"{self.scale} {tmp}"

    def __repr__(self):
        tmp = "{%s}" % ", ".join(
            [f"'{key}': {value}" for key, value in sorted(self._d.items())]
        )
        return f"<ParserHelper({self.scale}, {tmp})>"

    def __mul__(self, other):
        if isinstance(other, str):
            new = self.add(other, self._one)
        elif isinstance(other, Number):
            new = self.copy()
            new.scale *= other
        elif isinstance(other, self.__class__):
            new = self.operate(other.items())
            new.scale *= other.scale
        else:
            new = self.operate(other.items())
        return new

    __rmul__ = __mul__

    def __pow__(self, other):
        d = self._d.copy()
        for key in self._d:
            d[key] *= other
        return self.__class__(self.scale**other, d, non_int_type=self._non_int_type)

    def __truediv__(self, other):
        if isinstance(other, str):
            new = self.add(other, -1)
        elif isinstance(other, Number):
            new = self.copy()
            new.scale /= other
        elif isinstance(other, self.__class__):
            new = self.operate(other.items(), operator.sub)
            new.scale /= other.scale
        else:
            new = self.operate(other.items(), operator.sub)
        return new

    __floordiv__ = __truediv__

    def __rtruediv__(self, other):
        new = self.__pow__(-1)
        if isinstance(other, str):
            new = new.add(other, self._one)
        elif isinstance(other, Number):
            new.scale *= other
        elif isinstance(other, self.__class__):
            new = self.operate(other.items(), operator.add)
            new.scale *= other.scale
        else:
            new = new.operate(other.items(), operator.add)
        return new


#: List of regex substitution pairs.
_subs_re_list = [
    ("\N{DEGREE SIGN}", "degree"),
    (r"([\w\.\-\+\*\\\^])\s+", r"\1 "),  # merge multiple spaces
    (r"({}) squared", r"\1**2"),  # Handle square and cube
    (r"({}) cubed", r"\1**3"),
    (r"cubic ({})", r"\1**3"),
    (r"square ({})", r"\1**2"),
    (r"sq ({})", r"\1**2"),
    (
        r"\b([0-9]+\.?[0-9]*)(?=[e|E][a-zA-Z]|[a-df-zA-DF-Z])",
        r"\1*",
    ),  # Handle numberLetter for multiplication
    (r"([\w\.\)])\s+(?=[\w\(])", r"\1*"),  # Handle space for multiplication
]

#: Compiles the regex and replace {} by a regex that matches an identifier.
_subs_re = [
    (re.compile(a.format(r"[_a-zA-Z][_a-zA-Z0-9]*")), b) for a, b in _subs_re_list
]
_pretty_table = str.maketrans("⁰¹²³⁴⁵⁶⁷⁸⁹·⁻", "0123456789*-")
_pretty_exp_re = re.compile(r"(⁻?[⁰¹²³⁴⁵⁶⁷⁸⁹]+(?:\.[⁰¹²³⁴⁵⁶⁷⁸⁹]*)?)")


def string_preprocessor(input_string: str) -> str:
    input_string = input_string.replace(",", "")
    input_string = input_string.replace(" per ", "/")

    for a, b in _subs_re:
        input_string = a.sub(b, input_string)

    input_string = _pretty_exp_re.sub(r"**(\1)", input_string)
    # Replace pretty format characters
    input_string = input_string.translate(_pretty_table)

    # Handle caret exponentiation
    input_string = input_string.replace("^", "**")
    return input_string


def _is_dim(name: str) -> bool:
    return name[0] == "[" and name[-1] == "]"


class SharedRegistryObject:
    """Base class for object keeping a reference to the registree.

    Such object are for now Quantity and Unit, in a number of places it is
    that an object from this class has a '_units' attribute.

    Parameters
    ----------

    Returns
    -------

    """

    _REGISTRY: ClassVar[UnitRegistry]
    _units: UnitsContainer

    def __new__(cls, *args, **kwargs):
        inst = object.__new__(cls)
        if not hasattr(cls, "_REGISTRY"):
            # Base class, not subclasses dynamically by
            # UnitRegistry._init_dynamic_classes
            from . import application_registry

            inst._REGISTRY = application_registry.get()
        return inst

    def _check(self, other: Any) -> bool:
        """Check if the other object use a registry and if so that it is the
        same registry.

        Parameters
        ----------
        other

        Returns
        -------
        bool

        Raises
        ------
        ValueError
            if other don't use the same unit registry.
        """
        if self._REGISTRY is getattr(other, "_REGISTRY", None):
            return True

        elif isinstance(other, SharedRegistryObject):
            mess = "Cannot operate with {} and {} of different registries."
            raise ValueError(
                mess.format(self.__class__.__name__, other.__class__.__name__)
            )
        else:
            return False


class PrettyIPython:
    """Mixin to add pretty-printers for IPython"""

    default_format: str

    def _repr_html_(self) -> str:
        if "~" in self.default_format:
            return f"{self:~H}"
        return f"{self:H}"

    def _repr_latex_(self) -> str:
        if "~" in self.default_format:
            return f"${self:~L}$"
        return f"${self:L}$"

    def _repr_pretty_(self, p, cycle: bool):
        if "~" in self.default_format:
            p.text(f"{self:~P}")
        else:
            p.text(f"{self:P}")


def to_units_container(
    unit_like: QuantityOrUnitLike, registry: Optional[UnitRegistry] = None
) -> UnitsContainer:
    """Convert a unit compatible type to a UnitsContainer.

    Parameters
    ----------
    unit_like
        Quantity or Unit to infer the plain units from.
    registry
        If provided, uses the registry's UnitsContainer and parse_unit_name.  If None,
        uses the registry attached to unit_like.

    Returns
    -------
    UnitsContainer
    """
    mro = type(unit_like).mro()
    if UnitsContainer in mro:
        return unit_like
    elif SharedRegistryObject in mro:
        return unit_like._units
    elif str in mro:
        if registry:
            # TODO: document how to whether to lift preprocessing loop out to caller
            for p in registry.preprocessors:
                unit_like = p(unit_like)
            # TODO: Why not parse.units here?
            return registry._parse_units(unit_like)
        else:
            return ParserHelper.from_string(unit_like)
    elif dict in mro:
        if registry:
            return registry.UnitsContainer(unit_like)
        else:
            return UnitsContainer(unit_like)


def infer_base_unit(
    unit_like: QuantityOrUnitLike, registry: Optional[UnitRegistry] = None
) -> UnitsContainer:
    """
    Given a Quantity or UnitLike, give the UnitsContainer for it's plain units.

    Parameters
    ----------
    unit_like
        Quantity or Unit to infer the plain units from.
    registry
        If provided, uses the registry's UnitsContainer and parse_unit_name.  If None,
        uses the registry attached to unit_like.

    Returns
    -------
    UnitsContainer

    Raises
    ------
    ValueError
        The unit_like did not reference a registry, and no registry was provided.

    """
    d = udict()

    original_units = to_units_container(unit_like, registry)

    if registry is None and hasattr(unit_like, "_REGISTRY"):
        registry = unit_like._REGISTRY
    if registry is None:
        raise ValueError("No registry provided.")

    for unit_name, power in original_units.items():
        candidates = registry.parse_unit_name(unit_name)
        assert len(candidates) == 1
        _, base_unit, _ = candidates[0]
        d[base_unit] += power

    # remove values that resulted in a power of 0
    nonzero_dict = {k: v for k, v in d.items() if v != 0}

    return registry.UnitsContainer(nonzero_dict)


def getattr_maybe_raise(obj: Any, item: str):
    """Helper function invoked at start of all overridden ``__getattr__``.

    Raise AttributeError if the user tries to ask for a _ or __ attribute,
    *unless* it is immediately followed by a number, to enable units
    encompassing constants, such as ``L / _100km``.

    Parameters
    ----------
    item
        attribute to be found.

    Raises
    ------
    AttributeError
    """
    # Double-underscore attributes are tricky to detect because they are
    # automatically prefixed with the class name - which may be a subclass of obj
    if (
        item.endswith("__")
        or len(item.lstrip("_")) == 0
        or (item.startswith("_") and not item.lstrip("_")[0].isdigit())
    ):
        raise AttributeError(f"{obj!r} object has no attribute {item!r}")


def iterable(y: Any) -> bool:
    """Check whether or not an object can be iterated over."""
    try:
        iter(y)
    except TypeError:
        return False
    return True


def sized(y: Any) -> bool:
    """Check whether or not an object has a defined length."""
    try:
        len(y)
    except TypeError:
        return False
    return True


def create_class_with_registry(
    registry: UnitRegistry, base_class: type[TT]
) -> type[TT]:
    """Create new class inheriting from base_class and
    filling _REGISTRY class attribute with an actual instanced registry.
    """

    class_body = {
        "__module__": "pint",
        "_REGISTRY": registry,
    }

    return types.new_class(
        base_class.__name__,
        bases=(base_class,),
        exec_body=lambda ns: ns.update(class_body),
    )



from __future__ import annotations

import pathlib
import typing as ty
from typing import Optional, Union

from ..._vendor import flexcache as fc
from ..._vendor import flexparser as fp
from ..base_defparser import ParserConfig
from . import block, common, context, defaults, group, plain, system


class PintRootBlock(
    fp.RootBlock[
        ty.Union[
            plain.CommentDefinition,
            common.ImportDefinition,
            context.ContextDefinition,
            defaults.DefaultsDefinition,
            system.SystemDefinition,
            group.GroupDefinition,
            plain.AliasDefinition,
            plain.DerivedDimensionDefinition,
            plain.DimensionDefinition,
            plain.PrefixDefinition,
            plain.UnitDefinition,
        ],
        ParserConfig,
    ]
):
    body: fp.Multi[
        ty.Union[
            plain.CommentDefinition,
            common.ImportDefinition,
            context.ContextDefinition,
            defaults.DefaultsDefinition,
            system.SystemDefinition,
            group.GroupDefinition,
            plain.AliasDefinition,
            plain.DerivedDimensionDefinition,
            plain.DimensionDefinition,
            plain.PrefixDefinition,
            plain.UnitDefinition,
        ]
    ]


class PintSource(fp.ParsedSource[PintRootBlock, ParserConfig]):
    """Source code in Pint."""


class HashTuple(tuple):
    pass


class _PintParser(fp.Parser[PintRootBlock, ParserConfig]):
    """Parser for the original Pint definition file, with cache."""

    _delimiters = {
        "#": (
            fp.DelimiterInclude.SPLIT_BEFORE,
            fp.DelimiterAction.CAPTURE_NEXT_TIL_EOL,
        ),
        **fp.SPLIT_EOL,
    }
    _root_block_class = PintRootBlock
    _strip_spaces = True

    _diskcache: fc.DiskCache

    def __init__(self, config: ParserConfig, *args, **kwargs):
        self._diskcache = kwargs.pop("diskcache", None)
        super().__init__(config, *args, **kwargs)

    def parse_file(self, path: pathlib.Path) -> PintSource:
        if self._diskcache is None:
            return super().parse_file(path)
        content, basename = self._diskcache.load(path, super().parse_file)
        return content


class DefParser:
    skip_classes: tuple[type, ...] = (
        fp.BOF,
        fp.BOR,
        fp.BOS,
        fp.EOS,
        plain.CommentDefinition,
    )

    def __init__(self, default_config, diskcache):
        self._default_config = default_config
        self._diskcache = diskcache

    def iter_parsed_project(self, parsed_project: fp.ParsedProject):
        last_location = None
        for stmt in parsed_project.iter_blocks():
            if isinstance(stmt, fp.BOF):
                last_location = str(stmt.path)
            elif isinstance(stmt, fp.BOR):
                last_location = (
                    f"[package: {stmt.package}, resource: {stmt.resource_name}]"
                )

            if isinstance(stmt, self.skip_classes):
                continue

            if isinstance(stmt, common.DefinitionSyntaxError):
                # TODO: check why this assert fails
                # assert isinstance(last_location, str)
                stmt.set_location(last_location)
                raise stmt
            elif isinstance(stmt, block.DirectiveBlock):
                for exc in stmt.errors:
                    exc = common.DefinitionSyntaxError(str(exc))
                    exc.set_position(*stmt.get_position())
                    exc.set_raw(
                        (stmt.opening.raw or "") + " [...] " + (stmt.closing.raw or "")
                    )
                    exc.set_location(last_location)
                    raise exc

                try:
                    yield stmt.derive_definition()
                except Exception as exc:
                    exc = common.DefinitionSyntaxError(str(exc))
                    exc.set_position(*stmt.get_position())
                    exc.set_raw(stmt.opening.raw + " [...] " + stmt.closing.raw)
                    exc.set_location(last_location)
                    raise exc
            else:
                yield stmt

    def parse_file(
        self, filename: Union[pathlib.Path, str], cfg: Optional[ParserConfig] = None
    ):
        return fp.parse(
            filename,
            _PintParser,
            cfg or self._default_config,
            diskcache=self._diskcache,
            strip_spaces=True,
            delimiters=_PintParser._delimiters,
        )

    def parse_string(self, content: str, cfg: Optional[ParserConfig] = None):
        return fp.parse_bytes(
            content.encode("utf-8"),
            _PintParser,
            cfg or self._default_config,
            diskcache=self._diskcache,
            strip_spaces=True,
            delimiters=_PintParser._delimiters,
        )



import tokenize
from pint.pint_eval import _plain_tokenizer, uncertainty_tokenizer

tokenizer = _plain_tokenizer

input_lines = [
    "( 8.0 + / - 4.0 ) e6 m",
    "( 8.0 ± 4.0 ) e6 m",
    "( 8.0 + / - 4.0 ) e-6 m",
    "( nan + / - 0 ) e6 m",
    "( nan ± 4.0 ) m",
    "8.0 + / - 4.0 m",
    "8.0 ± 4.0 m",
    "8.0(4)m",
    "8.0(.4)m",
    "8.0(-4)m",  # error!
    "pint == wonderfulness ^ N + - + / - * ± m J s",
]

for line in input_lines:
    result = []
    g = list(uncertainty_tokenizer(line))  # tokenize the string
    for toknum, tokval, _, _, _ in g:
        result.append((toknum, tokval))

    print("====")
    print(f"input line: {line}")
    print(result)
    print(tokenize.untokenize(result))



#!/usr/bin/env python3

"""
    pint-convert
    ~~~~~~~~~~~~

    :copyright: 2020 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import argparse
import contextlib
import re

from pint import UnitRegistry

parser = argparse.ArgumentParser(description="Unit converter.", usage=argparse.SUPPRESS)
parser.add_argument(
    "-s",
    "--system",
    metavar="sys",
    default="SI",
    help="unit system to convert to (default: SI)",
)
parser.add_argument(
    "-p",
    "--prec",
    metavar="n",
    type=int,
    default=12,
    help="number of maximum significant figures (default: 12)",
)
parser.add_argument(
    "-u",
    "--prec-unc",
    metavar="n",
    type=int,
    default=2,
    help="number of maximum uncertainty digits (default: 2)",
)
parser.add_argument(
    "-U",
    "--no-unc",
    dest="unc",
    action="store_false",
    help="ignore uncertainties in constants",
)
parser.add_argument(
    "-C",
    "--no-corr",
    dest="corr",
    action="store_false",
    help="ignore correlations between constants",
)
parser.add_argument(
    "fr", metavar="from", type=str, help="unit or quantity to convert from"
)
parser.add_argument("to", type=str, nargs="?", help="unit to convert to")
try:
    args = parser.parse_args()
except SystemExit:
    parser.print_help()
    raise

ureg = UnitRegistry()
ureg.auto_reduce_dimensions = True
ureg.autoconvert_offset_to_baseunit = True
ureg.enable_contexts("Gau", "ESU", "sp", "energy", "boltzmann")
ureg.default_system = args.system


def _set(key: str, value):
    obj = ureg._units[key].converter
    object.__setattr__(obj, "scale", value)


if args.unc:
    import uncertainties

    # Measured constants subject to correlation
    #  R_i: Rydberg constant
    #  g_e: Electron g factor
    #  m_u: Atomic mass constant
    #  m_e: Electron mass
    #  m_p: Proton mass
    #  m_n: Neutron mass
    R_i = (ureg._units["R_inf"].converter.scale, 0.0000000000021e7)
    g_e = (ureg._units["g_e"].converter.scale, 0.00000000000035)
    m_u = (ureg._units["m_u"].converter.scale, 0.00000000050e-27)
    m_e = (ureg._units["m_e"].converter.scale, 0.00000000028e-30)
    m_p = (ureg._units["m_p"].converter.scale, 0.00000000051e-27)
    m_n = (ureg._units["m_n"].converter.scale, 0.00000000095e-27)
    if args.corr:
        # Correlation matrix between measured constants (to be completed below)
        #          R_i       g_e       m_u       m_e       m_p       m_n
        corr = [
            [1.0, -0.00206, 0.00369, 0.00436, 0.00194, 0.00233],  # R_i
            [-0.00206, 1.0, 0.99029, 0.99490, 0.97560, 0.52445],  # g_e
            [0.00369, 0.99029, 1.0, 0.99536, 0.98516, 0.52959],  # m_u
            [0.00436, 0.99490, 0.99536, 1.0, 0.98058, 0.52714],  # m_e
            [0.00194, 0.97560, 0.98516, 0.98058, 1.0, 0.51521],  # m_p
            [0.00233, 0.52445, 0.52959, 0.52714, 0.51521, 1.0],
        ]  # m_n
        (R_i, g_e, m_u, m_e, m_p, m_n) = uncertainties.correlated_values_norm(
            [R_i, g_e, m_u, m_e, m_p, m_n], corr
        )
    else:
        R_i = uncertainties.ufloat(*R_i)
        g_e = uncertainties.ufloat(*g_e)
        m_u = uncertainties.ufloat(*m_u)
        m_e = uncertainties.ufloat(*m_e)
        m_p = uncertainties.ufloat(*m_p)
        m_n = uncertainties.ufloat(*m_n)

    _set("R_inf", R_i)
    _set("g_e", g_e)
    _set("m_u", m_u)
    _set("m_e", m_e)
    _set("m_p", m_p)
    _set("m_n", m_n)

    # Measured constants with zero correlation
    _set(
        "gravitational_constant",
        uncertainties.ufloat(
            ureg._units["gravitational_constant"].converter.scale, 0.00015e-11
        ),
    )

    _set(
        "d_220",
        uncertainties.ufloat(ureg._units["d_220"].converter.scale, 0.000000032e-10),
    )

    _set(
        "K_alpha_Cu_d_220",
        uncertainties.ufloat(
            ureg._units["K_alpha_Cu_d_220"].converter.scale, 0.00000022
        ),
    )

    _set(
        "K_alpha_Mo_d_220",
        uncertainties.ufloat(
            ureg._units["K_alpha_Mo_d_220"].converter.scale, 0.00000019
        ),
    )

    _set(
        "K_alpha_W_d_220",
        uncertainties.ufloat(
            ureg._units["K_alpha_W_d_220"].converter.scale, 0.000000098
        ),
    )

    ureg._root_units_cache = {}
    ureg._build_cache()


def convert(u_from, u_to=None, unc=None, factor=None):
    q = ureg.Quantity(u_from)
    fmt = f".{args.prec}g"
    if unc:
        q = q.plus_minus(unc)
    if u_to:
        nq = q.to(u_to)
    else:
        nq = q.to_base_units()
    if factor:
        q *= ureg.Quantity(factor)
        nq *= ureg.Quantity(factor).to_base_units()
    prec_unc = use_unc(nq.magnitude, fmt, args.prec_unc)
    if prec_unc > 0:
        fmt = f".{prec_unc}uS"
    else:
        with contextlib.suppress(Exception):
            nq = nq.magnitude.n * nq.units

    fmt = "{:" + fmt + "} {:~P}"
    print(("{:} = " + fmt).format(q, nq.magnitude, nq.units))


def use_unc(num, fmt, prec_unc):
    unc = 0
    with contextlib.suppress(Exception):
        if isinstance(num, uncertainties.UFloat):
            full = ("{:" + fmt + "}").format(num)
            unc = re.search(r"\+/-[0.]*([\d.]*)", full).group(1)
            unc = len(unc.replace(".", ""))

    return max(0, min(prec_unc, unc))


def main():
    convert(args.fr, args.to)


if __name__ == "__main__":
    main()
